import java.util.*;
import java.io.*;
import java.text.DecimalFormat;
public class HuffmanEncoder
{
    public TreeNode tree; //for setting huffnums
    public ProbList probl; //list of chars and their probs
    public PQueue p; //priority queue
    private int depth; //like Inception, how deep you have gone
    String encodedFile; //will be a lot of 1's and 0's
    public DecimalFormat fourDigits = new DecimalFormat("0.0000"); //restricts to four decimal places
    
    //Huffman method does all work
    public void Huffman(String inputFile){
       probl = new ProbList(inputFile);
       probl.setListValues();
       //probl.printList(); used for debug 
       p = new PQueue();
       p.buildList(probl.probs, probl.listSize);
       //p.printList(); used for debug
       tree = huffman(probl.listSize);
       insertRep(tree, probl.probs, probl.listSize, "");
       probl.printList();
       encodedFile = encode(probl.file);
       //System.out.println(probl.file); used for debug
       System.out.println();
       System.out.println("Encoded File as a String: ");
       System.out.println(encodedFile);
       System.out.println();
       System.out.println("Overall Compression Value: " + fourDigits.format(getCompressionValue(probl.file, encodedFile)) + "% of original file size");
        
    }
    
    //encodes file into its Huffman representation
    public String encode(String file){
        String returnFile = "";
        for(int i = 0; i < file.length(); i++){
            int loc = probl.findNextChar(file.charAt(i));
            if(loc == -1){   //checks for errors
                System.out.println("Error encoding: can't find: " + file.charAt(i));
                System.exit(0);
            }
            returnFile = returnFile + probl.probs[loc].huffNum;
        }
        return returnFile;
        }
    
    //builds the tree based off probabilities
    public TreeNode huffman(int n){
        TreeNode tree = null;
        for(int i = 0; i < n-1; i++){
            tree = new TreeNode();
            tree.left = p.getLeast();
            tree.left.numSteps = i+1;
            tree.right = p.getLeast();
            tree.right.numSteps = i+1;
            tree.nodeProb = tree.left.nodeProb+tree.right.nodeProb;
            tree.huffNum = "";
            p.insert(tree);
        }
        return tree;
    }



   // creates representations for all nodes
   public void insertRep(TreeNode tree, Character chars[], int n, String repr) {
   //recursively sets values for each node
      String s1, s2;
      tree.huffNum = repr;
      if ((tree.left) == null && (tree.right) == null) {
         for (int i = 0; i < n; i++) 
            if (tree.theChar == chars[i].theChar)
               chars[i].huffNum = tree.huffNum;
         return;
      }
      s1 = repr;
      s1 += "0";
      insertRep(tree.left, chars, n, s1); // recursive call to the left
      s2 = repr;
      s2 += "1";
      insertRep(tree.right, chars, n, s2); // recursive call to the right
   }
   
   //gets the value for overall data compression
   public double getCompressionValue(String originalString, String huffString){
       double beginSize = 8 * originalString.length();
       double huffSize = huffString.length();
       double sizeDifference = (huffSize/beginSize) * 100;
       return sizeDifference; //how much space you saved
    }
    
    //sets up so you simply type in file
    //please note that the file extension must be entered correctly
   public static void main(String[] args){
       HuffmanEncoder e = new HuffmanEncoder();
       Scanner s = new Scanner(System.in);
       System.out.println("Please Enter File Name: ");
       String theFile = s.nextLine();
       e.Huffman(theFile);
    }

}
